# c36classes.py : 5 roues, le module appele par les interfaces (textes, ....)
# derive de m209.py v3 05/12/09, mais utilise les "OVERLAP"

import string
import random

class Wheel:
    def __init__(self,alphabet,shift):
        self.alphabet = alphabet
        self.lg = len(alphabet)
        self.pins = []
        for i in range(self.lg):
            self.pins.append( 0 )
        self.position = 0
        self.shiftWindow = shift

    def setPinsON(self,car):
        #i = string.index(self.alphabet,car)
        i = self.alphabet.index(car)
        self.pins[i] = 1

    def setPinsOFF(self,car):
        i = string.index(self.alphabet,car)
        self.pins[i] = 0
        
    def getAPins(self,car):
        i = string.index(self.alphabet,car)
        return self.pins[i]

    def getShift(self):
        return self.shiftWindow

    def turnClockwise(self):
        self.position = (self.position + 1 ) % self.lg

    def turnCounterClockwise(self):
        self.position = (self.position + self.lg - 1 ) % self.lg 

    def getLg( self):
        return len( self.alphabet )

    def getPosition(self):
        return self.position

    def setPosition(self, newPosition ):
        self.position = newPosition

    def getExternalKey(self):
        return self.alphabet[ self.position ]

    def getLetters(self):
        letters = []
        indice = self.position
        letters.append( self.alphabet[ indice -1 ] )
        letters.append( self.alphabet[ indice    ] )
        letters.append( self.alphabet[ (indice +1) % len(self.alphabet) ] )
        return letters

    def getExternalShiftedKey(self):
        return self.alphabet[(self.position + self.shiftWindow)%self.lg]

    def setInternalKey(self, lesPins ):
        for i in range(self.lg):
            if lesPins[i] == "1":
                self.pins[i] = 1
            else:
                self.pins[i] = 0

    def getInternalKey(self):
        ch = ""
        for i in range(self.lg):
            if self.pins[i] :
                ch = ch + self.alphabet[i]
            else:
                ch = ch + "_"
        return ch

    def isActive(self):
        pos_active = ( self.position + self.shiftWindow ) % self.lg
        if self.pins[ pos_active ] != 0:
            return 1
        else:
            return 0

    def zero(self):
        for i in range(self.lg):
            self.pins[i] = 0

    def info(self):
        print("Lenght       : ", self.lg)
        print("Alphabet     : ", self.alphabet)
        print("Shitf        : ", self.shiftWindow)
        print("Pins         : ", self.pins)
        print("Internal Key : ", self.getInternalKey())
        print("External key : ", self.getExternalKey())
        print("Position: ", self.position, " Actif : ", self.isActive())


class c36:
    def __init__(self, lettre_decalage, modele, debug):
        #self.decalage = string.index(string.ascii_uppercase,
        #    lettre_decalage)
        self.decalage = (string.ascii_uppercase).index( lettre_decalage )
        self.modele = modele
        self.debug = debug
        self.wheels = []   
        # ecarts entre alignement AAAAA et Pins actifs: JIHHG
        self.wheels.append( Wheel("ABCDEFGHIJKLMNOPQRSTUVXYZ",9  ) )
        self.wheels.append( Wheel("ABCDEFGHIJKLMNOPQRSTUVX",8) )
        self.wheels.append( Wheel("ABCDEFGHIJKLMNOPQRSTU",7  ) )        
        self.wheels.append( Wheel("ABCDEFGHIJKLMNOPQRS",7) )
        self.wheels.append( Wheel("ABCDEFGHIJKLMNOPQ",6  ) )
        self.nbOfWheels = len( self.wheels )
        self.lugs = []
        if  self.modele == "c35":
            self.lugs = [1,2,4,8,10]
        elif self.modele == "c36":
            self.lugs = [1,2,4,6,12]
        elif self.modele == "over":
            self.lugs = [1,2,3,7,14]
        elif ":" in self.modele :
            tb = self.modele.split(":")
            self.lugs = list(map(int,tb))
        self.modeCode  = 1
        self.counter  = 0
        self.history = ""

    def getHistory(self):
        return self.history

    def setLetterShift(self, lettre_decalage):
        self.decalage = string.ascii_uppercase.index(lettre_decalage)

    def setInternalKey(self, pinsKey):
        self.wheels[0].setInternalKey( pinsKey[0:25] )
        self.wheels[1].setInternalKey( pinsKey[25:(25+23)] )
        self.wheels[2].setInternalKey( pinsKey[(25+23):(25+23+21)] )
        self.wheels[3].setInternalKey( pinsKey[(25+23+21):(25+23+21+19)] )
        self.wheels[4].setInternalKey( pinsKey[(25+23+21+19):] )

    def getLetterShift(self):
        return string.ascii_uppercase[ self.decalage ]

    def setLugs(self,lesLugs):
        self.lugs = lesLugs

    def getLugs(self):
        return self.lugs

    def setPinsON(self,laroue,char):
        self.wheels[laroue].setPinsON(char)

    def setPinsOFF(self,laroue,char):
        self.wheels[laroue].setPinsOFF(char)

    def getPins(self,laroue,char):
        return self.wheels[laroue].getAPins(char)

    def stepAWheel(self, aWheel ):
        self.wheels[ aWheel ].turnClockwise()

    def step(self):
        for i in range( self.nbOfWheels ):        
            self.wheels[i].turnClockwise()
        self.counter = self.counter + 1

    def stepBack(self):
        for i in range( self.nbOfWheels ):        
            self.wheels[i].turnCounterClockwise()
        self.counter = self.counter - 1
        if ( self.counter < 0 ):
            self.counter = 999

    def rewind(self):
        while ( self.counter != 0 ):
            self.stepBack()

    def getPosition( self ):
        return self.counter

    def setPosition( self, posi ):
        self.counter = posi

    def getKeyOfAWheel( self, aWheel ):
        return self.wheels[ aWheel ].getExternalKey()

    def getKeyShiftedOfAWheel( self, aWheel ):
        return self.wheels[ aWheel ].getExternalShiftedKey()

    def getLetters( self, aWheel ):
        return self.wheels[ aWheel ].getLetters()

    def getExternalKey( self ):
        ch = ""
        for i in range( self.nbOfWheels ):
            ch = ch + self.getKeyOfAWheel( i )
        return ch

    def getExternalShiftedKey( self ):
        ch = ""
        for i in range( self.nbOfWheels ):
            ch = ch + self.getKeyShiftedOfAWheel( i )
        return ch

    def avanceRoue( self, roue ):
        self.wheels[roue].turnClockwise()

    def setExternalKey( self, key ):
        for i in range( len(key) ):
            non_trouve = 1
            lettre_cle = key[i]
            lg = self.wheels[i].getLg()
            for j in range (lg):
                if lettre_cle == self.wheels[i].getExternalKey():
                    non_trouve = 0                    
                    break
                self.avanceRoue(i)
            if non_trouve: 
                print("**** erreur cle externe **** "); exit

    def getPinsOfAWheel( self, aWheel ):
        return self.wheels[ aWheel ].getInternalKey()

    def reinit(self):
        for i in range( self.nbOfWheels ):
            self.wheels[i].setPosition(0)
        self.counter = 0
        self.history = ""

    def zero(self):
        for i in range( self.nbOfWheels ):        
            self.wheels[i].zero()
        for i in range( self.nbOfWheels ):
            self.lugs[ i ] = 0
        self.setZeroOver()
        self.setLetterShift( 'Z' )
        self.reinit()

    def zeroPins(self):
        for i in range( self.nbOfWheels ):        
            self.wheels[i].zero()
        if self.debug:
            print("ZERO PINS")

    def cipher(self,char):
        index = string.ascii_uppercase.index(char)
        
        activesPins = []
        for j in range( self.nbOfWheels ):
            if self.wheels[ j ].isActive():
                activesPins.append( 1 )
            else:
                activesPins.append( 0 )
        shift = 0
        for j in range( self.nbOfWheels ):
            shift = shift + activesPins[j] * self.lugs[j]
        if self.modele == "over":
            if activesPins[0] != 0 and activesPins[4] != 0 : shift -= 1
            if activesPins[1] != 0 and activesPins[4] != 0 : shift -= 1
        j = ( 26 + self.decalage - index + shift ) % 26 
        char2 = string.ascii_uppercase[ j ]
        if ( self.debug ):
            ch = "Counter: %03d Key: %6s %6s %2d " % (self.getPosition(),self.getExternalKey(), \
                self.getExternalShiftedKey(), self.decalage)
            ch = ch + ("[ %d %d %d %d %d ]" % tuple(activesPins))
            tdecl = (self.decalage + shift ) % 26
            ch = ch + ("%2d %2d %c %c" % (shift, tdecl, char, char2))
            self.history = self.history + ch + '\n'
            print(ch)
        self.step()
        return char2

    def getNbOfBars(self):
        return self.nbOfBars

    def getNbOfWheels(self):
        return self.nbOfWheels

    def toggleModeCode( self):
        if self.modeCode: self.modeCode = 0
        else: self.modeCode = 1

    def toggleDebug( self ):
        if self.debug: self.debug = 0
        else: self.debug = 1

    def getModeCode( self):
        return self.modeCode

    def info(self):
        print("Mode : ", end=' ') 
        if self.modeCode: print("CODE")
        else: print("DECODE")
        for i in range( self.nbOfWheels ):
            print(i+1, ":", self.wheels[i].getInternalKey())
            if ( self.debug ): self.wheels[i].info()
        print("Lugs    : ", self.getLugs())
        print("Shift Letter: ", self.getLetterShift())
        print("Actives Pins : ", end=' ') 
        for i in range( self.nbOfWheels ):
            print(self.wheels[i].isActive(), end=' ')
        print("Counter: ", self.getPosition(), "Key     : ", self.getExternalKey())
#================================================
if __name__ == "__main__" :
    c = c36("Z","over",0)
    calage = string.ascii_uppercase[ random.randint(0,25) ]
    c.setLetterShift( calage )
    c.setInternalKey(   "00110011001111001" +
                        "1100110011001111001" +
                        "010100110011001111001" +
                        "11111100110011001111001" +
                        "0011001100110011001111001")
    plain = "BONJOUR"
    crypto = ""
    for car in plain:
        crypto += c.cipher( car )
    print("Crypto: ", crypto)
    c.reinit()
    plain = ""
    for car in crypto:
        plain += c.cipher( car )
    print("Plain:  ", plain)
    c.info()
